/* 
    3D Engine

    2005 Jason Hunt
    justinsanemsu@hotmail.com

    file: 3dmath.cpp
*/

#include <math.h>
#include <stdlib.h>

#include "3dmath.h"
#include "types.h"


/* point addition */
point3d operator+(point3d lhs, point3d rhs)
{
    point3d t;

    t.x = lhs.x + rhs.x;
    t.y = lhs.y + rhs.y;
    t.z = lhs.z + rhs.z;

    return(t);
}
/* end of point3d operator+ */


/* point subtraction */
point3d operator-(point3d lhs, point3d rhs)
{
    point3d t;

    t.x = lhs.x - rhs.x;
    t.y = lhs.y - rhs.y;
    t.z = lhs.z - rhs.z;

    return(t);
}
/* end of point3d operator- */



/* point addition */
point3d operator+=(point3d &lhs, point3d rhs)
{
    lhs.x += rhs.x;
    lhs.y += rhs.y;
    lhs.z += rhs.z;

    return(lhs);
}
/* end of point3d operator+ */


/* point addition */
point3d operator-=(point3d &lhs, point3d rhs)
{
    lhs.x -= rhs.x;
    lhs.y -= rhs.y;
    lhs.z -= rhs.z;

    return(lhs);
}
/* end of point3d operator+ */


/* scale * */
point3d operator*=(point3d &lhs, double rhs)
{
    lhs.x *= rhs;
    lhs.y *= rhs;
    lhs.z *= rhs;
 
    return(lhs);
}
/* end scaler multiply */



/* scale * */
point3d operator*(point3d lhs, double rhs)
{
    point3d t;

    t.x = lhs.x * rhs;
    t.y = lhs.y * rhs;
    t.z = lhs.z * rhs;
 
    return(t);
}
/* end scaler multiply */



/* scale * */
point3d operator/=(point3d &lhs, double rhs)
{
    if (rhs == 0)
    {
        lhs = get_point(0,0,0);
        return(lhs);
    }

    lhs.x /= rhs;
    lhs.y /= rhs;
    lhs.z /= rhs;
 
    return(lhs);
}
/* end scaler multiply */


/* scale * */
point3d operator/(point3d lhs, double rhs)
{
    point3d t;

    if (rhs == 0)
    {
        lhs = get_point(0,0,0);
        return(lhs);
    }

    t.x = lhs.x / rhs;
    t.y = lhs.y / rhs;
    t.z = lhs.z / rhs;
 
    return(t);
}
/* end scaler multiply */


/* absolute value */
point3d abs(point3d p)
{
    point3d r;
    r.x = fabs(p.x);
    r.y = fabs(p.y);
    r.z = fabs(p.z);
    return(r);
}
/* end of absolute value */


/* conpute dot product */
float dot_product(point3d a, point3d b)
{
    return(a.x*b.x + a.y*b.y + a.z*b.z);
}
/* end of dot_product */


/* compute lenght */
float length(point3d v)
{
    return(sqrt(dot_product(v, v)));
}
/* end of length */


/* compute cross product */
point3d cross_product(point3d v1, point3d v2)
{
    point3d c;

    c.x = v1.y * v2.z - v2.y * v1.z;
    c.y = v1.z * v2.x - v2.z * v1.x;
    c.z = v1.x * v2.y - v2.x * v1.y;

    return(c);
}
/* end of cross_product */


/* normalize a vector */
void normalize(point3d &p)
{
    double mag = sqrt(p.x * p.x + p.y * p.y + p.z * p.z);
    
    if (mag == 0)
    {
        p = get_point(0, 0, 0);
        return;
    }

    // normalize
    p /= mag;
}
/* end of normalize */


/* compute a surface normal */
point3d get_normal(point3d p1, point3d p2, point3d p3)
{
    point3d n = cross_product(p3 - p1, p2 - p1);

    normalize(n);
 
    return(n);
}
/* end of get_normal */


/* returns a point */
point3d get_point(float x, float y, float z)
{
    point3d p;

    p.x = x;
    p.y = y;
    p.z = z;
    
    return(p);
}
/* end of get_point */


/* update trig date for a direction */
void update_trig(trig_type &trig, point3d &dir)
{
    trig.cos_x = cos(dir.x);   trig.sin_x = sin(dir.x);
    trig.cos_y = cos(dir.y);   trig.sin_y = sin(dir.y);
    trig.cos_z = cos(dir.z);   trig.sin_z = sin(dir.z);
}
/* end of update_trig */


/* return light modification */
float calculate_light(point3d pos, point3d source, point3d normal)
{
    pos -= source;

    normalize(pos);

    return(128.0 * dot_product(pos, normal));
}
/* end of calculate_light */


/* return a direction vector from trig data */
point3d get_direction_vector( trig_type trig )
{
    point3d t;

    t.x = trig.sin_y * trig.cos_x;
    t.z = trig.cos_y * trig.cos_x;
    t.y = -trig.sin_x;

    normalize(t);

    return(t);
}
/* end of get_direction_vector */



/*
    End of 3d operations
*/





/*
    Operations on 2d vectors
*/


/* point addition */
point2d operator+(point2d lhs, point2d rhs)
{
    point2d t;

    t.x = lhs.x + rhs.x;
    t.y = lhs.y + rhs.y;

    return(t);
}
/* end of point3d operator+ */


/* point subtraction */
point2d operator-(point2d lhs, point2d rhs)
{
    point2d t;

    t.x = lhs.x - rhs.x;
    t.y = lhs.y - rhs.y;

    return(t);
}
/* end of point3d operator- */



/* point addition */
point2d operator+=(point2d &lhs, point2d rhs)
{
    lhs.x += rhs.x;
    lhs.y += rhs.y;

    return(lhs);
}
/* end of point3d operator+ */


/* point addition */
point2d operator-=(point2d &lhs, point2d rhs)
{
    lhs.x -= rhs.x;
    lhs.y -= rhs.y;

    return(lhs);
}
/* end of point3d operator+ */


/* scale * */
point2d operator*=(point2d &lhs, double rhs)
{
    lhs.x *= rhs;
    lhs.y *= rhs;
 
    return(lhs);
}
/* end scaler multiply */


/* scale * */
point2d operator*(point2d lhs, double rhs)
{
    point2d t;

    t.x = lhs.x * rhs;
    t.y = lhs.y * rhs;
 
    return(t);
}
/* end scaler multiply */



/* scale * */
point2d operator/=(point2d &lhs, double rhs)
{
    if (rhs == 0)
    {
        lhs = get_point(0,0);
        return(lhs);
    }

    lhs.x /= rhs;
    lhs.y /= rhs;
 
    return(lhs);
}
/* end scaler multiply */


/* scale * */
point2d operator/(point2d lhs, double rhs)
{
    point2d t;

    if (rhs == 0)
    {
        lhs = get_point(0,0);
        return(lhs);
    }

    t.x = lhs.x / rhs;
    t.y = lhs.y / rhs;
 
    return(t);
}
/* end scaler multiply */


/* absolute value */
point2d abs(point2d p)
{
    point2d r;
    r.x = fabs(p.x);
    r.y = fabs(p.y);
    return(r);
}
/* end of absolute value */


/* conpute dot product */
float dot_product(point2d a, point2d b)
{
    return(a.x*b.x + a.y*b.y);
}
/* end of dot_product */


/* compute lenght */
float length(point2d v)
{
    return(sqrt(dot_product(v, v)));
}
/* end of length */


/* returns a point */
point2d get_point(float x, float y)
{
    point2d p;

    p.x = x;
    p.y = y;
    
    return(p);
}
/* end of get_point */


/* return a floating point value in the interval */
float f_rand(float start, float end)
{
    return((end - start) * (rand() / (float) RAND_MAX) + start);
}
/* end of f_rand */


