/* 
    3D Engine

    2005 Jason Hunt
    justinsanemsu@hotmail.com

    file: cerate.cpp
*/

#include <math.h>

#include "world.h"
#include "debug.h"
#include "3dmath.h"

/* returns a 3d cord from componets */
inline vertex_class get_cord(float x, float y, float z)
{
    vertex_class t;
    t.base.x = x, t.base.y = y, t.base.z = z;
    return(t);
}
/* end of get_cord */


/* creats a trianlge from index data */
inline triangle_class get_triangle(vertex_class *v_data,
                                   unsigned int p0, unsigned int p1, unsigned int p2,
                                   unsigned char r, unsigned char g, unsigned char b, int t_num,
                                   int invert)

{
    triangle_class t;

    t.tx_num = t_num;

    t.vertex[0] = &v_data[p0];
    t.vertex[1] = &v_data[p1];
    t.vertex[2] = &v_data[p2];

    t.color.r = r;    t.color.g = g;    t.color.b = b;

    t.shade = true;

    t.mutate_dir = 0;

    // set texture cords
    if (invert)
    {   // triangle upside down?
        t.t_cords[0] = get_point(1, 0);
        t.t_cords[1] = get_point(1, 1);
        t.t_cords[2] = get_point(0, 0);
    } else
    {
        t.t_cords[0] = get_point(0, 1);
        t.t_cords[1] = get_point(0, 0);
        t.t_cords[2] = get_point(1, 1);
    }

    t.t_map = true; // enable texture mapping
    return(t);
}
/* end of get_triangle */


/* adds a cube to the world */
void world_class::add_cube( float size, unsigned int tx_num, bool shade )
{
    if (num_object >= max_object) reallocate_object_memory();

    objects[num_object] = (object_class*) malloc(sizeof(object_class));

    object_class *t = objects[num_object];

    t->num_vertex = 8;
    t->vertex_data = (vertex_class *)malloc(t->num_vertex * sizeof(vertex_class));

    if (!t->vertex_data)
    {
        t->num_vertex = 0;
        add_debug("world_class::add_cube", "Unable to get memory for vertex data");
        return;
    }

    t->vertex_data[0] = get_cord(-size,  size, -size);
    t->vertex_data[1] = get_cord( size,  size, -size);
    t->vertex_data[2] = get_cord( size, -size, -size);
    t->vertex_data[3] = get_cord(-size, -size, -size);
    t->vertex_data[4] = get_cord(-size,  size,  size);
    t->vertex_data[5] = get_cord( size,  size,  size);
    t->vertex_data[6] = get_cord( size, -size,  size);
    t->vertex_data[7] = get_cord(-size, -size,  size);

    t->num_triangle = 12;
    t->triangle_data = (triangle_class *) malloc(t->num_triangle * sizeof(triangle_class));

    if (!t->triangle_data)
    {
        t->num_triangle = 0;
        add_debug("world_class::add_cube", "Unable to get memory for triangle data");
        return;
    }

    t->triangle_data[0]  = get_triangle(t->vertex_data, 1, 2, 0, 0,  0, 128, tx_num, 0);
    t->triangle_data[1]  = get_triangle(t->vertex_data, 3, 0, 2, 0,  0, 128, tx_num, 1);
    t->triangle_data[2]  = get_triangle(t->vertex_data, 1, 5, 2, 0,  0, 128, tx_num, 0);
    t->triangle_data[3]  = get_triangle(t->vertex_data, 6, 2, 5, 0,  0, 128, tx_num, 1);
    t->triangle_data[4]  = get_triangle(t->vertex_data, 1, 0, 5, 0,  0, 128, tx_num, 0);
    t->triangle_data[5]  = get_triangle(t->vertex_data, 4, 5, 0, 0,  0, 128, tx_num, 1);
    t->triangle_data[6]  = get_triangle(t->vertex_data, 3, 7, 0, 0,  0, 128, tx_num, 0);
    t->triangle_data[7]  = get_triangle(t->vertex_data, 4, 0, 7, 0,  0, 128, tx_num, 1);
    t->triangle_data[8]  = get_triangle(t->vertex_data, 2, 6, 3, 0,  0, 128, tx_num, 0);
    t->triangle_data[9]  = get_triangle(t->vertex_data, 7, 3, 6, 0,  0, 128, tx_num, 1);
    t->triangle_data[10] = get_triangle(t->vertex_data, 5, 4, 6, 0,  0, 128, tx_num, 0);
    t->triangle_data[11] = get_triangle(t->vertex_data, 7, 6, 4, 0,  0, 128, tx_num, 1);


    for (int i = 0; i < 12; i++)
        t->triangle_data[i].shade = shade;

    // clear direction
    t->position = get_point(0,0,0);
    t->position_rate = get_point(0,0,0);

    t->direction = get_point(0,0,0);
    t->direction_rate = get_point(0,0,0);

    t->radius = sqrt(size*size);
    t->shade = true;


    update_trig(t->trig, t->direction);

    num_object++;
}
/* end of add cube */


/* adds a sphere to the world */
void world_class::add_sphere( unsigned int points, bool shade, unsigned int tx_num)
{
    add_debug(" world_class::add_sphere", ""); 

    if (num_object >= max_object) reallocate_object_memory();

    objects[num_object] = (object_class*) malloc(sizeof(object_class));

    object_class *t = objects[num_object];

    unsigned int rings = points;

    t->num_vertex = points * rings;
    t->vertex_data = (vertex_class *)malloc(t->num_vertex * sizeof(vertex_class));

    float y = 0.05;

    for (int r = 0; r < rings; r++)
    {
        float  m = 0.1 * sin(r/(float)(rings-1)  * M_PI); 
        float  y = 0.1 * cos(r/(float)(rings-1)  * M_PI); 

        for (int i = 0; i < points; i++)
        {
            float z = m*sin(i/(float)points * 2 * M_PI);
            float x = m*cos(i/(float)points * 2 * M_PI);

            t->vertex_data[i + r*points] = get_cord(x, y, z);
        }
    }

    t->num_triangle = points*2*(rings-1);

    t->triangle_data = (triangle_class *) malloc(t->num_triangle * sizeof(triangle_class));

    for (int r = 0; r < rings-1; r++)
    {
        unsigned char c = 64 + rand()%64;

        unsigned int base = r * points;
        
        for (int i = 0; i < points; i++)
        {
            t->triangle_data[i + 2*base] = get_triangle(t->vertex_data,               // base data
                                                      i + base,                 // first index
                                                      (i+1) % points + base ,   // index
                                                      i + points + base ,       // index
                                                      0,  0, 120, tx_num,0);       // color

            t->triangle_data[i + 2*base].shade = shade;
        }

        for (int i = 0; i < points; i++)
        {
            t->triangle_data[i+points+2*base] = get_triangle(t->vertex_data,
                                                            points + (i+1)%points+base,
                                                            i + points+base, 
                                                            (i+1)%points+base, 
                                                            0, 0, 60, tx_num,1);

             // copy color from the other triangles in this face
             t->triangle_data[i+points+2*base].color =  t->triangle_data[i+2*base].color;
             t->triangle_data[i+points+2*base].shade = shade;
        }
    }

    // clear direction

    t->position = get_point(0,0,0);
    t->position_rate = get_point(0,0,0);

    t->direction = get_point(0,0,0);
    t->direction_rate = get_point(0,0,0);

    t->radius = 0.1;

    t->shade = true;
    update_trig(t->trig, t->direction);
    num_object++;
}
/* end of add_sphere */



/* adds a cube to the world */
void world_class::add_test( void )
{
    if (num_object >= max_object) reallocate_object_memory();

    objects[num_object] = (object_class*) malloc(sizeof(object_class));

    object_class *t = objects[num_object];

    t->num_vertex = 3;
    t->vertex_data = (vertex_class *)malloc(t->num_vertex * sizeof(vertex_class));

    if (!t->vertex_data)
    {
        t->num_vertex = 0;
        add_debug("world_class::add_cube", "Unable to get memory for vertex data");
        return;
    }

    t->vertex_data[0] = get_cord(-0.1,  0.1, -0.1);
    t->vertex_data[1] = get_cord( 0.1,  0.1, -0.1);
    t->vertex_data[2] = get_cord( -0.1, -0.1, -0.1);
 
    t->num_triangle = 1;
    t->triangle_data = (triangle_class *) malloc(t->num_triangle * sizeof(triangle_class));

    if (!t->triangle_data)
    {
        t->num_triangle = 0;
        add_debug("world_class::add_cube", "Unable to get memory for triangle data");
        return;
    }

    t->triangle_data[0]  = get_triangle(t->vertex_data, 0, 1, 2, 0,  0, 128, 0,0);
    // clear direction

    t->position = get_point(0,0,0);
    t->position_rate = get_point(0,0,0);

    t->direction = get_point(0,0,0);
    t->direction_rate = get_point(0,0,0);

    t->radius = 0.14;
    t->shade = true;
    update_trig(t->trig, t->direction);

    num_object++;
}
/* end of add cube */



void world_class::make_textures( void )
{
    color_type c;

    for (int y = 0; y < texture_size; y++)
        for (int x = 0; x < texture_size; x++)
        {
            c.r = 110 + rand() % 32;
            c.g = c.r;
            c.b = c.g;
            texture[0][x][y] = c;

            c.r = 110 + rand() % 32;
            c.g = 0;
            c.b = 0;
            texture[1][x][y] = c;

            c.r = 0;
            c.g = 110 + rand() % 32;
            c.b = 0;
            texture[2][x][y] = c;

            c.r = 0;
            c.g = 0;
            c.b = 110 + rand() % 32;
            texture[3][x][y] = c;

            c.r = 110 + rand() % 32;
            c.g = c.r;
            c.b = 0;
            texture[4][x][y] = c;
        }
}


