/* 
    3D Engine

    2005 Jason Hunt
    justinsanemsu@hotmail.com

    file: object.cpp
*/


#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "world.h"
#include "object.h"
#include "types.h"
#include "debug.h"
#include "3dmath.h"
#include "const.h"

extern world_class world; // gloable engine data 
extern bool paused;


/* loads an object from a file */
bool object_class::load_from_file(char *filename)
{
    add_debug("0bject_class::load_from_file", filename);

    num_vertex = 0;
    num_triangle = 0;

    FILE *f = fopen(filename, "rt");

    // is the file there?
    if (!f)
    {
        add_debug("0bject_class::load_from_file", "File not found");
        return(false);
    }

    char buffer[500];

    // start loading the file

    // position x, y, z

    fscanf(f, "%lf %lf %lf\n", &position.x, &position.y, &position.z);

    sprintf(buffer, "Position (%f, %f, %f)", position.x, position.y, position.z);
//    add_debug("0bject_class::load_from_file", buffer);

    fscanf(f, "%d\n", &num_vertex);    // number of verticies

    sprintf(buffer, "Vertex number: %d", num_vertex);
    add_debug("0bject_class::load_from_file", buffer);

    // Get vertex memory
    vertex_data = (vertex_class *) malloc(num_vertex * sizeof(vertex_class));
    
    // make sure it's allocated
    if (vertex_data == NULL)
    {
        add_debug("0bject_class::load_from_file", "Unable to get vertex memory");
        return(false);
    }

    // load vertex data 
    for (unsigned int i = 0; i < num_vertex && !feof(f); i++)
    {
        fscanf(f, "%lf %lf %lf\n", &vertex_data[i].base.x, 
                                   &vertex_data[i].base.y, 
                                   &vertex_data[i].base.z);
    }

    // add to debug log if needed
    for (unsigned int i = 0; i < num_vertex; i++)
    {
        sprintf(buffer, "(%f %f %f)", vertex_data[i].base.x,  
                                      vertex_data[i].base.y, 
                                      vertex_data[i].base.z);

        add_debug("   vertex", buffer);
    }

    // number of triangles

    fscanf(f, "%d\n", &num_triangle);

    sprintf(buffer, "triangle number: %d", num_triangle);
    add_debug("0bject_class::load_from_file", buffer);

    // get triangle memory

    triangle_data = (triangle_class *) malloc(num_triangle * sizeof(triangle_class));

    if (triangle_data == NULL)
    {
        add_debug("0bject_class::load_from_file", "Unable to get triangle memory");
        return(false);
    }

    // triangle data
    for (unsigned int i = 0; i < num_triangle && !feof(f); i++)
    {
        unsigned int tri[3]; // index to triangles

        unsigned int color[3];

        // get the triangle indexs
        fscanf(f, "%d %d %d %d %d %d\n", &tri[0], &tri[1], &tri[2], 
                                         &color[0], &color[1],&color[2]);

        triangle_data[i].color.r = color[0];
        triangle_data[i].color.g = color[1];
        triangle_data[i].color.b = color[2];
        triangle_data[i].shade = false;

        // set pointers
        triangle_data[i].vertex[0] = &vertex_data[tri[0]];
        triangle_data[i].vertex[1] = &vertex_data[tri[1]];
        triangle_data[i].vertex[2] = &vertex_data[tri[2]];

        triangle_data[i].mutate_dir = 0;
    }


    // clear variables
    direction.x = 0;      direction.y = 0;      direction.z = 0;
    direction_rate.x = 0; direction_rate.y = 0; direction_rate.z = 0;
    position_rate.x = 0;  position_rate.y = 0;  position_rate.z = 0;

    radius = 0.2;


    shade = true;

    update_trig(trig, direction);

    fclose(f);

    return(true);
}
/* end of load_object_from_file */


/* free object memory */
void object_class::unload( void )
{
    if (vertex_data != NULL) free(vertex_data);
    if (triangle_data != NULL) free(triangle_data);
}
/* end of unload */



/* preform rotation */
void inline rotate(float &s, float &t, float sn, float cs)
{
    float ns, nt;

    ns = s * cs - t * sn;
    nt = s * sn + t * cs;

  	s = ns; t = nt;
}
/* end of rotate */


/* transform into 3 space */
void object_class::transform( camera_class camera )
{
    vertex_class *v = vertex_data;

    for (unsigned int i = 0; i < num_vertex; i++, v++)
    {
        // assume it's clipped
        v->clip = true;

        // load point data from baseline object data
        point3d p = v->base;

        // rotate around self;
        // rotate the point arround the y dir
        rotate(p.x, p.z, trig.sin_y, trig.cos_y);   // rotate the point arround the y dir
        rotate(p.z, p.y, trig.sin_x, trig.cos_x);   // rotate the point arround the x dir
        rotate(p.x, p.y, trig.sin_z, trig.cos_z);   // rotate the point arround the z dir

        // save cords in world space
        v->self = p + position;

        // get vector to center of object
        v->center_vec =position- v->self;
        normalize(v->center_vec);

        // center around position and camera
        p += position - camera.position;
        
        // rotate around camera
        rotate(p.x, p.z, camera.trig.sin_y, camera.trig.cos_y); // rotate the point arround the y dir
        rotate(p.z, p.y, camera.trig.sin_x, camera.trig.cos_x); // rotate the point arround the x dir

        // save rotated points for later
        v->rot = p;

        // now we are in view space

        if (p.z > 0.0001) // front clip
        {
            float z_inv = 1.0 / p.z;

            // translate to the screen
            v->screen_x = (display_x / 2) + (int)(depth_x * p.x * z_inv);
            v->screen_y = (display_y / 2) + (int)(depth_y * p.y * z_inv); 
         
            // viewport clip
            if (v->screen_x >= 0 && v->screen_x < display_x &&
                v->screen_y >= 0 && v->screen_y < display_y)
                    v->clip = false;
        }
     }
}
/* end of transsform */


/* does anything occupy the space the object is about to enter? */
bool object_class::okay_to_move(point3d p)
{
    // check against other objects 
    for (int i = 0; i < world.num_object; i++)
        if (world.objects[i] != this)
            if (length(p - world.objects[i]->position) < world.objects[i]->radius + radius) return(false);

    return(true);
}
/* end of okay_to_move */


/* move the object */
void object_class::move( void )
{
    if (paused) return; 

    if (length(position_rate) < 0.001) return;

    point3d next_pos = position + position_rate;
 
    if ( okay_to_move(next_pos))
    {
        position = next_pos;
    }
    else
    {
        position_rate *= -1;
        direction_rate *= -1;
    }

    // inside bounding shere?
    if (dot_product(position, position) > 6)
    {
        position_rate *= -1;

  //      normalize(position);
//        position *= 2;

    }

    direction += direction_rate;

    update_trig(trig, direction);
}
/* end of move */


/* draw object to the buffer */
void object_class::draw( void )
{
    for (unsigned int tri = 0; tri < num_triangle; tri++)
        triangle_data[tri].draw();
}
/* end of draw_object */


/* return a random color */
color_type random_color( void )
{
    color_type t;
    t.r = 0;//
    t.g = 0;//64 + rand() % 128;
    t.b = 64 + rand() % 128;
    return(t);
}
/* end of random color */


/* randomize_placement */
void object_class::randomize_placement( void )
{
    position = get_point(f_rand(-1.0, 1.0), f_rand(-1.0, 1.0), f_rand(-1.0, 1.0));

    // make sure we can land there
    while (!okay_to_move(position))
        position = get_point(f_rand(-1.0, 1.0), f_rand(-1.0, 1.0), f_rand(-1.0, 1.0));

    direction_rate = get_point(f_rand(-0.1, 0.1), f_rand(-0.1, 0.1), f_rand(-0.1, 0.1));
    position_rate = get_point(f_rand(-0.01, 0.01), f_rand(-0.01, 0.01), f_rand(-0.01, 0.01));
}
/* end of randomize_placement */


/* set color of entire object */
void object_class::set_color(unsigned char r, unsigned char g, unsigned char b)
{
    for (int t = 0; t < num_triangle; t++)
    {
        triangle_data[t].color.r = r;
        triangle_data[t].color.g = g;
        triangle_data[t].color.b = b;
    }
}
/* end of set_color */

