/* 
    3D Engine

    2005 Jason Hunt
    justinsanemsu@hotmail.com

    file: triangle.cpp
*/


#include <stdio.h>
#include <math.h>

#include "world.h"
#include "triangle.h"
#include "display.h"
#include "debug.h"
#include "3dmath.h"

extern bool shader;

extern unsigned int z_buffer[display_x*display_y];

extern world_class world;


#define fixed_num   8*65536
#define fixed_shift  19


#define to_fixed(x) (x * 65536)
#define to_int(x)   (x >> 16)
#define fixed_mul(x,y) (((x) * (y)) >> 16)
#define fixed_div(x,y) ((y<<16) / x))


/* swaps two floats */
void inline swap(float &a, float &b)
{
	float tmp;
	tmp = a;
	a = b;
	b = tmp;
}
/* end of swap */

/* this is for debugging, writes a pixel to the buffer */
/*void put_pixel(int x, int y, color_type color)
{
     if (x >= 0 && x < display_x && y >= 0 && y < display_y)
        buffer[x + y *display_x] = color;
}*/
/* end of put pixel */


typedef struct edge_type
{
    // start and end of span on screen in screen cords
    int x_start, x_end;                                

    // start and end of light intensity
    float l_start, l_end;                              

    // start and end positions in texture space
    float uz_start, vz_start, z_start;
    float uz_end, vz_end, z_end;
};


/* fill edge buffer with values for the connecting line */
void calc_edge(edge_type *edge, vertex_class *v1, vertex_class *v2, 
                                point2d t_start, point2d t_end)
{
    if (v1->rot.z < 0) v1->rot.z = 0.001;
    if (v2->rot.z < 0) v2->rot.z = 0.001;

    float z1 = 1 / v1->rot.z;
    float z2 = 1 / v2->rot.z;

    float x1 = v1->screen_x;
    float y1 = v1->screen_y;
    float l1 = v1->light;
    point2d t1 = t_start * z1;

    float x2 = v2->screen_x;
    float y2 = v2->screen_y;
    float l2 = v2->light;
    point2d t2 = t_end * z2;

    point2d tx_grad;            // gradient in texture space

    // draw top down
    if (y2 < y1)
    {
        swap (y1, y2);
        swap (x1, x2);

        swap (l1, l2);

        swap (t1.x, t2.x);
        swap (t1.y, t2.y);

        swap (z1, z2);
    }

    // calculate graidents
            
    float dx,       // change in x for the vert change
          dl;       // change in light

    float dz;       // change in 1/z

    int height = y2 - y1;           // this is the height that the deltas are calculated with
                                    // is there is any cliping, this will be different

    if (height <= 0) return;

    float y_diff_inv =  1 / (double)(y2 - y1);    

    dx = (x2 - x1) * y_diff_inv;            // screen x gradient (1d)
    dl = (l2 - l1) * y_diff_inv;            // color graident (1d)
    dz = (z2 - z1) * y_diff_inv;

    tx_grad = (t2 - t1) * y_diff_inv;       // compute graident for world space (3d)

    // top clip
    if (y1 < 0)
    {
        // this is how much was clipped
        int top_diff = height - y2;

        y1 = 0;

        x1 += top_diff*dx;           // update screen x
        l1 += top_diff*dl;           // update light

        z1 += top_diff*dz;

        t1 += tx_grad*top_diff;  
    }

    // bottom clip
    if (y2 > display_y-1) y2 = display_y-1;

    // actual on screen height
    int draw_height = y2 - y1;

    if (draw_height <= 0) return;   // another paranoia check
      
    edge_type *e = edge + (int)y1;  // point to first base

    while (draw_height--)
    {   
        int x = (int)x1;
          
        if (x < e->x_start) 
        {
            e->x_start = x;
            e->l_start = (int)l1;
            e->uz_start = t1.x;
            e->vz_start = t1.y;
            e->z_start = z1;
        }

        if (x > e->x_end) 
        {
            e->x_end = x;
            e->l_end = (int)l1;
            e->uz_end = t1.x;
            e->vz_end = t1.y;
            e->z_end = z1;
        }

        x1 += dx;           // update screen x
        l1 += dl;           // update light

        z1 += dz;

        t1 += tx_grad;      // update point in texture space

        e++;                // next edge
    }   

}
/* end of calc_edge */


/* draw a texture mapped triangle */
void triangle_class::draw(void)
{
    if (shade)
    {
        vertex[0]->light = calculate_light(vertex[0]->self, world.light_source, vertex[0]->center_vec);
        vertex[1]->light = calculate_light(vertex[1]->self, world.light_source, vertex[1]->center_vec);
        vertex[2]->light = calculate_light(vertex[2]->self, world.light_source, vertex[2]->center_vec);
    }
    else
    {
        vertex[0]->light = 0;
        vertex[1]->light = 0;
        vertex[2]->light = 0;
    }

    edge_type edge[display_y];  // data for left and right edges

    int y_min = display_y, y_max = -1; // set to impossiable values

    // calculate top and bottom points
    for (unsigned int i = 0; i < 3; i++)
    {
        if (vertex[i]->screen_y < y_min) y_min = vertex[i]->screen_y;
        if (vertex[i]->screen_y > y_max) y_max = vertex[i]->screen_y;
    }

    // clip
    if (y_min < 0) y_min = 0;
    if (y_max > display_y - 1) y_max = display_y - 1;

    // sanity check
    if (y_min > y_max) return;

    // reset to impossible values
    for (int i = y_min; i <= y_max; i++)
    {   
        edge[i].x_start = display_x;
        edge[i].x_end = -1;
    }

    // fill edge buffer
    calc_edge(edge, vertex[0], vertex[1], t_cords[0], t_cords[1]);
    calc_edge(edge, vertex[1], vertex[2], t_cords[1], t_cords[2]);
    calc_edge(edge, vertex[2], vertex[0], t_cords[2], t_cords[0]);

    // pointer to first edge buffer
    edge_type *e = edge + y_min;

    // dump each row
    for (int scan_y = y_min; scan_y <= y_max; scan_y++)
    {
        int start = e->x_start;
        int end = e->x_end;

        int tx_len = end - start; // length of span in pixels before clip

        int clip_advance = 0;

        // left clip and interpolation correction
        if (start < 0)
        {
            clip_advance = -start;
            start = 0;
        }

        // right clip
        if (end >= display_x) 
            end = display_x - 1;
    
        // sanity check
        if (end > start)
        {                
            // create pointer to start screen point
            color_type *p = buffer + scan_y * display_x + start;
            unsigned int *zb = z_buffer + scan_y * display_x + start;     

            // calculate deltas

            int inv_tx_len = fixed_num / tx_len;

            // linear interpolation for u/z, v/z and 1/z
            int dl =  (int)((e->l_end  - e->l_start)  * inv_tx_len);    // delta for light
            int duz = (int)((e->uz_end - e->uz_start) * inv_tx_len);
            int dvz = (int)((e->vz_end - e->vz_start) * inv_tx_len);
            int dz =  (int)((e->z_end  - e->z_start)  * inv_tx_len);

            // skip over left clipped region
            // these are
            int l =  (int)(e->l_start  * fixed_num) +  dl * clip_advance;
            int uz = (int)(e->uz_start * fixed_num) + duz * clip_advance;
            int vz = (int)(e->vz_start * fixed_num) + dvz * clip_advance;
            int z =  (int)(e->z_start  * fixed_num)  + dz * clip_advance;

            // scale to size of texture
            uz *= texture_size;
            vz *= texture_size;

            duz *= texture_size;
            dvz *= texture_size;

            int len = end - start; // length of scanline

            #define i_len (32)
            
            int iu = (long long)(uz)*fixed_num/z; // uz*fixed_num/z;
            int iv = (long long)(vz)*fixed_num/z; // uz*fixed_num/z;

            // deltas for linear interpolation
            int diu = ((long long)(uz + i_len * duz)*fixed_num / (z + i_len * dz) - iu) /i_len;
            int div = ((long long)(vz + i_len * dvz)*fixed_num / (z + i_len * dz) - iv) /i_len;

            unsigned int count = i_len;

            // copy span to buffer
            // this is the inner loop of the engine

            while (len-- && z > 0)
            {  
                if (count-- == 0)
                {
                    count = i_len;

                    iu = (long long int)uz*fixed_num/z; 
                    iv = (long long int)vz*fixed_num/z; 

                    // deltas for linear interpolation
                    diu = ((long long)(uz + i_len * duz)*fixed_num / (z + i_len * dz) - iu) /i_len;
                    div = ((long long)(vz + i_len * dvz)*fixed_num / (z + i_len * dz) - iv) /i_len;
               }

                unsigned int u = ((iu)>>fixed_shift) & 0x7f;
                unsigned int v = ((iv)>>fixed_shift) & 0x7f;

                // get texel 
               // color_type t =  world.texture[tx_num][(ru) & 0x7f][(rv) & 0x7f];
                color_type t =  world.texture[tx_num][u][v];

                int ls = l>>fixed_shift;

                int tmp = t.r + ls;
                if (tmp > 255) tmp = 255;
                if (tmp < 0) tmp = 0;
                t.r = tmp;

                tmp = t.g + ls;
                if (tmp > 255) tmp = 255;
                if (tmp < 0) tmp = 0;
                t.g = tmp;

                tmp = t.b + ls;
                if (tmp > 255) tmp = 255;
                if (tmp < 0) tmp = 0;
                t.b = tmp;

                // closer in z_buffer?
                if (z > *zb)
                {
                    *p = t;                // write to display buffer
                    *zb = z;                // set new 1/z
                }

                p++;                // next display buffer
                zb++;               // next z_buffer

                l += dl;            // light

                // these are interpolated
                iu += diu;
                iv += div;

                // these are needed for recalulations
                uz += duz;
                vz += dvz;
                z += dz;
            }
        }
        
        // next edge
        e++;
    }
}
/* end of triangle::draw */



/* returns true if at least some of the trianlge is on the screen */
bool triangle_class::is_visible( void )
{
	return ((vertex[2]->rot.x * ((vertex[0]->rot.z * vertex[1]->rot.y)-(vertex[0]->rot.y*vertex[1]->rot.z))+
     	     vertex[2]->rot.y * ((vertex[0]->rot.x * vertex[1]->rot.z)-(vertex[0]->rot.z*vertex[1]->rot.x))+
    	     vertex[2]->rot.z * ((vertex[0]->rot.y * vertex[1]->rot.x)-(vertex[0]->rot.x*vertex[1]->rot.y)) > 0) &&
          !(vertex[0]->clip && vertex[1]->clip && vertex[2]->clip));
}
/* end of is_visible */



/* sets the average z value for later sorting */
void triangle_class::compute_values( void )
{
    average_z = (vertex[0]->rot.z + vertex[1]->rot.z + vertex[2]->rot.z) / 3.0;

    normal = get_normal(vertex[0]->self, vertex[1]->self, vertex[2]->self);

    point3d center;

    center.x = (vertex[0]->self.x + vertex[1]->self.x + vertex[2]->self.x) / 3.0;
    center.y = (vertex[0]->self.y + vertex[1]->self.y + vertex[2]->self.y) / 3.0;
    center.z = (vertex[0]->self.z + vertex[1]->self.z + vertex[2]->self.z) / 3.0;

    light = calculate_light(center, world.light_source, normal);

//    vertex[0]->light = calculate_light(vertex[0]->self, world.light_source, normal);
//    vertex[1]->light = calculate_light(vertex[1]->self, world.light_source, normal);
//    vertex[2]->light = calculate_light(vertex[2]->self, world.light_source, normal);
}
/* end of compute_z */



